use std::fmt::Debug;
use std::ops::{Add, AddAssign, Div, DivAssign, Mul, MulAssign, Sub, SubAssign};

/// Trait representing all numeric primitives and `Complex`.
pub trait Numeric:
Sized + Copy + Clone + Debug +
Add<Output=Self> + AddAssign + Sub<Output=Self> + SubAssign +
Mul<Output=Self> + MulAssign + Div<Output=Self> + DivAssign +
PartialEq {
    /// Function returning the whole number in the requested type.
    fn whole(value: u8) -> Self;

    /// Default tolerance for comparisons.
    fn epsilon() -> Self;

    /// Calculate the absolute value.
    fn abs(self) -> Self;

    /// Check for equality.
    /// For integrals, the tolerance is ignored.
    fn is_equal_to(self, other: &Self, tolerance: Self) -> bool;
}

impl Numeric for f32 {
    fn whole(value: u8) -> Self {
        value as Self
    }

    fn epsilon() -> Self {
        1e-9
    }

    fn abs(self) -> Self {
        self.abs()
    }

    fn is_equal_to(self, other: &Self, tolerance: Self) -> bool {
        (self - *other).abs() <= tolerance
    }
}

impl Numeric for f64 {
    fn whole(value: u8) -> Self {
        value as Self
    }

    fn epsilon() -> Self {
        1e-9
    }

    fn abs(self) -> Self {
        self.abs()
    }

    fn is_equal_to(self, other: &Self, tolerance: Self) -> bool {
        (self - *other).abs() <= tolerance
    }
}

impl Numeric for i8 {
    fn whole(value: u8) -> Self {
        value as Self
    }

    fn epsilon() -> Self {
        0
    }

    fn abs(self) -> Self {
        self.abs()
    }

    fn is_equal_to(self, other: &Self, tolerance: Self) -> bool {
        debug_assert_eq!(tolerance, 0);
        self == *other
    }
}

impl Numeric for i16 {
    fn whole(value: u8) -> Self {
        value as Self
    }

    fn epsilon() -> Self {
        0
    }

    fn abs(self) -> Self {
        self.abs()
    }

    fn is_equal_to(self, other: &Self, tolerance: Self) -> bool {
        debug_assert_eq!(tolerance, 0);
        self == *other
    }
}

impl Numeric for i32 {
    fn whole(value: u8) -> Self {
        value as Self
    }

    fn epsilon() -> Self {
        0
    }

    fn abs(self) -> Self {
        self.abs()
    }

    fn is_equal_to(self, other: &Self, tolerance: Self) -> bool {
        debug_assert_eq!(tolerance, 0);
        self == *other
    }
}

impl Numeric for i64 {
    fn whole(value: u8) -> Self {
        value as Self
    }

    fn epsilon() -> Self {
        0
    }

    fn abs(self) -> Self {
        self.abs()
    }

    fn is_equal_to(self, other: &Self, tolerance: Self) -> bool {
        debug_assert_eq!(tolerance, 0);
        self == *other
    }
}

impl Numeric for i128 {
    fn whole(value: u8) -> Self {
        value as Self
    }

    fn epsilon() -> Self {
        0
    }

    fn abs(self) -> Self {
        self.abs()
    }

    fn is_equal_to(self, other: &Self, tolerance: Self) -> bool {
        debug_assert_eq!(tolerance, 0);
        self == *other
    }
}

impl Numeric for isize {
    fn whole(value: u8) -> Self {
        value as Self
    }

    fn epsilon() -> Self {
        0
    }

    fn abs(self) -> Self {
        self.abs()
    }

    fn is_equal_to(self, other: &Self, tolerance: Self) -> bool {
        debug_assert_eq!(tolerance, 0);
        self == *other
    }
}

impl Numeric for u8 {
    fn whole(value: u8) -> Self {
        value as Self
    }

    fn epsilon() -> Self {
        0
    }

    fn abs(self) -> Self {
        self
    }

    fn is_equal_to(self, other: &Self, tolerance: Self) -> bool {
        debug_assert_eq!(tolerance, 0);
        self == *other
    }
}

impl Numeric for u16 {
    fn whole(value: u8) -> Self {
        value as Self
    }

    fn epsilon() -> Self {
        0
    }

    fn abs(self) -> Self {
        self
    }

    fn is_equal_to(self, other: &Self, tolerance: Self) -> bool {
        debug_assert_eq!(tolerance, 0);
        self == *other
    }
}

impl Numeric for u32 {
    fn whole(value: u8) -> Self {
        value as Self
    }

    fn epsilon() -> Self {
        0
    }

    fn abs(self) -> Self {
        self
    }

    fn is_equal_to(self, other: &Self, tolerance: Self) -> bool {
        debug_assert_eq!(tolerance, 0);
        self == *other
    }
}

impl Numeric for u64 {
    fn whole(value: u8) -> Self {
        value as Self
    }

    fn epsilon() -> Self {
        0
    }

    fn abs(self) -> Self {
        self
    }

    fn is_equal_to(self, other: &Self, tolerance: Self) -> bool {
        debug_assert_eq!(tolerance, 0);
        self == *other
    }
}

impl Numeric for u128 {
    fn whole(value: u8) -> Self {
        value as Self
    }

    fn epsilon() -> Self {
        0
    }

    fn abs(self) -> Self {
        self
    }

    fn is_equal_to(self, other: &Self, tolerance: Self) -> bool {
        debug_assert_eq!(tolerance, 0);
        self == *other
    }
}

impl Numeric for usize {
    fn whole(value: u8) -> Self {
        value as Self
    }

    fn epsilon() -> Self {
        0
    }

    fn abs(self) -> Self {
        self
    }

    fn is_equal_to(self, other: &Self, tolerance: Self) -> bool {
        debug_assert_eq!(tolerance, 0);
        self == *other
    }
}

#[cfg(test)]
mod tests {
    use crate::Numeric;

    #[test]
    fn floats_are_numeric() {
        assert_eq!(f32::whole(0), 0.0);
        assert_eq!(f32::whole(1), 1.0);
        assert_eq!(f64::whole(0), 0.0);
        assert_eq!(f64::whole(1), 1.0);
    }

    #[test]
    fn integers_are_numeric() {
        // not sure yet, if I want this to be possible
        assert_eq!(i8::whole(255), -1);

        assert_eq!(i8::whole(0), 0);
        assert_eq!(i8::whole(1), 1);
        assert_eq!(i16::whole(0), 0);
        assert_eq!(i16::whole(1), 1);
        assert_eq!(i32::whole(0), 0);
        assert_eq!(i32::whole(1), 1);
        assert_eq!(i64::whole(0), 0);
        assert_eq!(i64::whole(1), 1);
        assert_eq!(i128::whole(0), 0);
        assert_eq!(i128::whole(1), 1);
        assert_eq!(isize::whole(0), 0);
        assert_eq!(isize::whole(1), 1);

        assert_eq!(u8::whole(0), 0);
        assert_eq!(u8::whole(1), 1);
        assert_eq!(u16::whole(0), 0);
        assert_eq!(u16::whole(1), 1);
        assert_eq!(u32::whole(0), 0);
        assert_eq!(u32::whole(1), 1);
        assert_eq!(u64::whole(0), 0);
        assert_eq!(u64::whole(1), 1);
        assert_eq!(u128::whole(0), 0);
        assert_eq!(u128::whole(1), 1);
        assert_eq!(usize::whole(0), 0);
        assert_eq!(usize::whole(1), 1);
    }
}
